#!/usr/bin/env python
# -*- coding: utf-8 -*-

'''
This class will handle the current multi-network hard mesh objects.
'''

import maya.cmds as cmds


import hmScripts.hm_global_variables as nc
import hmScripts.Main.hmOperatorObject as netOb
import hmScripts.Main.hmNodeConnector as hmConn
import hmScripts.Main.hmBaseObject as hmNode
import hmScripts.Main.hmUtility as util


class HmNetworkHandler(object):

    # Attributes
    HILIGHTED_MESH_COLOR = 25
    HILIGHTED_CURVES_COLOR = 12


    # Internal options
    VERBOSE = 0 # 2 level of depth


    #===========================================================
    #==================== MAIN METHODS =========================
    #===========================================================
    def __init__(self):
        '''
        Initialize the class.

        Args:
            input_net_objs: a list of > HmOperatorObject <

        Returns:
            HmNetworkHandler

        '''
        # The network name of the object that initialize everything.
        # We use it for initializing properly the current_obj variable
        self.starting_net_name = None

        # Here we store the network objects
        self.net_objs = None

        # Current object ID that refers to the first list
        self.current_obj = 0


        # Options
        #
        self.local_visibility = None


    def initialize_from_selection(self, selected_obj, hm_type):
        '''
        Starting from a first input object, we go straight forward and discover the last network in chain.
        Then going backward we build a complete list of networks, even with branches!

        Args:
            selected_obj: One valid hm object
            hm_type: The type of hm object (we could remove that actually by checking it internally)

        Returns:
            None
        '''


        # ============== INTERNAL FUNCTIONS ==============

        def is_valid_out_mesh(input_object):
            '''
            Return if the provided hmNode describing an input mesh is valid for our backward network discovery.
            '''
            # We get the object type
            obj_type = hmConn.HmOperationConnect.get_object_type( input_object.get_name() )

            # If is hm_out_mesh
            if obj_type == nc.hm_out_mesh:
                # We check that has a real connection to a network
                if input_object.get_connected('inMesh', 'in', node_type = 'choice') or input_object.get_connected('inMesh', 'in', node_type = 'hmCompoundNode'):
                    return True
                # If not
                else:
                    return False





        # ============== START OF THE METHOD ==============

        # We initialize the current obj as HmOperatorObject
        first_operator = netOb.HmOperatorObject(selected_obj, hm_type)
        self.starting_net_name = first_operator.get_network_name()
        self.net_objs = [ first_operator ]


        # Print message
        if HmNetworkHandler.VERBOSE: print 'The Handler initialized from %s ' % self.starting_net_name


        # ============== Network chain search ==============
        # Get the last network for searching forward
        last_obj = self.get_last_operator()

        # Override in to the globals
        self.net_objs = [last_obj]

        # Networks list to query while going backward
        network_list = [last_obj]

        # <<<<<< moving backward
        while ( len(network_list) != 0 ):

            if HmNetworkHandler.VERBOSE:
                print 'la lista ha %s' % [network_list[x].get_network_name() for x in range(len(network_list))]

            # Get always the first element
            # Since we are removing 1 element every loop
            current_network = network_list[0]

            # We get the input meshes, and search for an output mesh type.
            # If we find there is a network going backward.
            in_meshes = current_network.get_input_meshes()

            if HmNetworkHandler.VERBOSE:
                print 'buono %s ha %s input mesh' % (current_network.get_network_name(), len(in_meshes) )

            while len(in_meshes) != 0:


                # Get always the first element
                # Since we are removing 1 element every loop
                current_mesh = in_meshes[0]

                if HmNetworkHandler.VERBOSE:print 'verifico %s' % current_mesh.get_name()

                # If is a valid hm_out_mesh
                if is_valid_out_mesh( current_mesh ):
                    # Create a new network instance
                    connected_net = netOb.HmOperatorObject( current_mesh.get_name(), nc.hm_out_mesh)
                    # Add to the list for the next query step
                    network_list.append( connected_net )
                    # Extend the list of current networks
                    self.net_objs.insert(0, connected_net)


                # Always remove after checking
                in_meshes.remove( current_mesh )

                if HmNetworkHandler.VERBOSE: print 'RIMUOVO %s' % current_mesh.get_name()


            # Always remove after checking
            network_list.remove( current_network )

            if HmNetworkHandler.VERBOSE: print 'RIMUOVO %s' % current_network.get_network_name()

        # ===Reverse the list for having the last operation at the top of the tree view
        self.net_objs.reverse()

        # Test if we are in global or local visibility mode
        self.__get_visibility_state_from_networks()

        # === Find the network that belongs to the initializer object
        for x,operator in enumerate(self.net_objs):
            if operator.get_network_name() == self.starting_net_name:
                self.set_current_operator(x)




        # Print message
        if self.VERBOSE: print 'The Handler stored %s ' % [self.net_objs[x].get_network_name() for x in range(len(self.net_objs))]



    #===========================================================
    #================== OPERATIONS METHODS =====================
    #===========================================================

    #==================== GENERAL =======================
    def get_operations_list(self):
        '''
        Return the list that contains all the HmOperatorObject related to the current multi network operation

        Args:
            None

        Returns:
            list[HmOperatorObject,HmOperatorObject, ...]

        '''
        return self.net_objs

    def get_last_operator(self):
        '''
        Return the last network object in a network chain by following
        all the connections going forward.

        Args:
            None

        Returns:
            HmOperatorObject
        '''
        # Safety options
        max_iter = 100
        curr_iter = 0

        # Main var to update in the while
        last_obj = self.net_objs[-1]

        # >>>>>> moving forward
        next_connection = True


        while (next_connection != None):

            curr_iter += 1

            # We get the result mesh, to query if
            # it is a source for another hm operation
            result_mesh = last_obj.get_result_mesh()
            shape = result_mesh.get_shape()
            out_data = shape.get_connected('outMesh', 'out')

            if HmNetworkHandler.VERBOSE: print 'get_last_operator', last_obj, result_mesh, shape, out_data

            # We check that the connection is going to an hm node
            if out_data:

                for connected in out_data:

                    # If it's going to an intersection or an offset node it means that is valid
                    if connected[0].is_type('pfxToon') or connected[0].is_type('hmOffsetCurveNode'):
                        # print 'inizializzo un nuovo network!', last_obj

                        # Create a new instance network for the next iteration
                        # and initialize as input mesh, for moving forward
                        last_obj = netOb.HmOperatorObject( result_mesh.get_name(), nc.hm_in_mesh)

            # If there are no out connections we stop the loop
            else:
                next_connection = None


            if curr_iter == max_iter:

                if self.VERBOSE: cmds.warning('Reached maximum search depth, the operator stack might be incomplete.')

                next_connection = None

        return last_obj

    def set_current_operator(self, input_ID):
        '''
        Set an operator as "current" for performing commands on it.

        Args:
            input_ID: The ID of the network do set as current

        Returns:
            None
        '''
        self.current_obj = input_ID

    def get_current_operator(self):
        '''
        Return the python object of the current network to access
        all the relative methods related to the operation.

        Args:
            None

        Returns:
            HmOperatorObject
        '''
        return self.net_objs[self.current_obj]

    def get_current_operator_ID(self):
        '''
        Return the Integer corresponding to the current active network
        in the self.net_objs[] list

        Args:
            None

        Returns:
            Int
        '''
        return self.current_obj

    def add_new_operation(self):
        '''
        When we are operating on a network and we create a new operation on one of the involved objects
        we add the operator directly in the network

        Args:
            None

        Returns:
            None
        '''
        pass

    def reset(self):
        '''
        Reset the internal storage
        '''
        self.net_objs = None
    #==================== EDITING =======================
    @util.undo
    def delete_operation(self, *args):
        '''
        Delete the current operator from the herarchy of networks.
        Reconnect properly the remaining elements to give a valid result.

        Args:
            None

        Returns:
            None
        '''
        # Get the current operator ID
        curr_ID = self.get_current_operator_ID()
        net_lenght = len(self.net_objs) -1

        if curr_ID == net_lenght and net_lenght > 0:
            raise Exception('You can\'t delete the first operator in chain.')
        else:
            # ================ DISCONNECT AND RECONNECT ==========================
            # Disconnect and remove from memory
            disconnected_operator = self.__connect_prev_with_next(curr_ID)[0]


            # ================ CLEAN MESHES ==================================
            # Clean the input meshes from the hm attributes and remove the drawing override
            meshes_to_clean = disconnected_operator.get_input_meshes()

            for mesh in meshes_to_clean:
                # delete the attributes only on input meshes
                if mesh.get_object_type() == nc.hm_in_mesh:
                    # Delete the custom attributes, the poly smooth, inter mesh and the offset
                    util.clean_object( mesh )


            # ================ DELETE CURRENT ==================================
            # Delete current network from scene
            disconnected_operator.get_network_object().delete()


            # ================ DISPLAY THE LAST OF CHAIN ==================================
            if len(self.net_objs) != 0:
                # Make sure that is visible
                last_result = self.net_objs[0].get_result_mesh()
                last_result.set_attribute('visibility', 1)

                # Unlock the result mesh if the inputs are hidden
                in_m_vis = self.net_objs[0].get_network_object().get_attribute(nc.base_meshes_vis_name)

                # print 'in mesh visibility %s for %s' % ( in_m_vis, self.net_objs[0] )

                # If we are in local visibility, with the input on, when we delete the operation
                # we will risk to have the out mesh locked. Make sure that this does not happen
                if not in_m_vis and self.local_visibility or in_m_vis == False:
                    last_result.set_attribute('overrideDisplayType', 0)
                    last_result.get_shape().set_attribute('overrideDisplayType', 0)

                # Force the update of the ui by changing selection.. is not clean do do it here
                cmds.select(last_result.get_name())
            else:
                cmds.select(cl=True)


    def bypass_operator(self, operatorID, value):
        '''
        Mute the current operator by bypassing its result or by turning it on again

        Args:
            operatorID: the int of the network in the self.net_objs list
            value: 0 for bypassing it, 1 for computing it

        Returns:
            None
        '''
        # Get the current operator to act on
        curr_operator = self.net_objs[operatorID]

        if value == 1:
            curr_operator.get_network_object().set_attribute(nc.operation_status_name, 0)
        if value == 0:
            curr_operator.get_network_object().set_attribute(nc.operation_status_name, 1)



    def move_operator_up(self, *values):
        '''
        By providing the operator name and an integer-position number we
        recreate the connections and update the UI according to the new order

        Args:
            operatorID: the int of the network to delete in the self.net_objs list
            new_position: the int the new position in self.net_objs list

        Returns:
            None
        '''
        NotImplemented


    def move_operator_down(self, *values):
        '''
        By providing the operator name and an integer-position number we
        recreate the connections and update the UI according to the new order

        Args:
            operatorID: the int of the network to delete in the self.net_objs list
            new_position: the int the new position in self.net_objs list

        Returns:
            None
        '''
        NotImplemented


    @util.undo
    def move_operator(self, curr_ID, new_ID):
        '''
        Write here what the function does.

        Args:
            curr_ID: The ID of the operator to move.
            new_ID: The ID rapresenting the new position in the stack.

        Returns:
            Status: bool: if true, update and refresh the UI, a valid operation has been performed, if false go back.
        '''
        def get_in_connections(operator_ID):
            """
            Given an operator index get its input
            connections by querying the preview operator out_mesh connections.
            """
            preview_operator = self.net_objs[operator_ID+1]

            # Store the preview mesh
            # Get the preview operator out mesh
            preview_result_mesh = preview_operator.get_result_mesh()

            # Store the out mesh connections
            transform_connections = preview_result_mesh.get_connected('worldMatrix[0]', 'out')
            shape_connections = preview_result_mesh.get_shape().get_connected('outMesh', 'out')

            return (transform_connections, shape_connections)


        result_status = True

        # Get the current operator ID
        curr_ID = int(curr_ID)
        new_ID = int(new_ID)

        # print 'move from %s to %s' % (curr_ID, new_ID)

        net_lenght = len(self.net_objs)


        # If is the first of the chain, and there is actually a chain, abort the action
        if curr_ID == net_lenght-1 and net_lenght > 0:
            result_status = False
            # raise Exception('You can\'t move the first operator in chain.')
        elif new_ID == net_lenght-1:
            result_status = False
            # raise Exception('You can\'t replace the first operator in chain.')
        # If not, we can go ahead
        else:

            self.set_compute_state(0)

            # ================ GET CONNECTION DATA ==========================
            moving_operator = self.net_objs[curr_ID]
            moved_in_connections = get_in_connections(curr_ID)


            # Disconnect the current operator from the networks
            #
            self.__connect_prev_with_next(curr_ID)


            # for x, obj in enumerate(self.net_objs):
            # 	print obj.get_network_name() + str(x) + '\n'


            # Now the memory has one operator less

            input_operator = self.net_objs[new_ID]
            in_hm_mesh = input_operator.get_result_mesh()
            out_operator = None

            # If the new position is in between two other operators
            if new_ID < net_lenght and new_ID > 0:
                out_operator = self.net_objs[new_ID-1]
                out_op_in_connections = get_in_connections(new_ID-1)




            # ================ SET CONNECTION DATA ==========================
            self.__connect_operators( in_hm_mesh, moved_in_connections )
            # print 'moved in <<<<', input_operator, '\n'
            for c in moved_in_connections:
                    for conn in c:
                     print conn

            if out_operator:
                # print 'moved out >>>>', out_operator, '\n'
                print moving_operator.get_result_mesh()
                # for c in out_op_in_connections:
                # 	for conn in c:
                # 	 print conn
                self.__connect_operators( moving_operator.get_result_mesh(), out_op_in_connections )








            # Insert in memory
            self.net_objs.insert(new_ID, moving_operator)



            # for x, obj in enumerate(self.net_objs):
            # 	print obj.get_network_name() + str(x) + '\n'




            # Update intersection curves
            #




            # ================ DISPLAY THE LAST OF CHAIN ==================================
            if len(self.net_objs) != 0:
                # Make sure that is visible
                last_result = self.net_objs[0].get_result_mesh()
                last_result.set_attribute('visibility', 1)

                # Hide the current one
                moving_operator.get_result_mesh().set_attribute('visibility', 0)
                # Force the update of the ui by changing selection.. is not clean do do it here
                cmds.select(last_result.get_name())
            else:
                cmds.select(cl=True)

            self.set_compute_state(1)

        return result_status


    #===========================================================
    #==================== GLOBAL METHODS =======================
    #===========================================================
    # Global methods acts on all the operators loaded
    def set_new_net_smooth(self, value):
        '''
        Increase or decrease the smooth level for the entire network

        Args:
            value: 0 for lowering on one level. 1 for increasinf of 1 level.

        Returns:
            None
        '''
        for operator in self.net_objs:
            # Get the current in meshes
            in_meshes = operator.get_input_meshes()

            for mesh in in_meshes:
                # Apply only to the clean 'in mesh' and not to result in mesh
                if mesh.get_object_type() == nc.hm_in_mesh:
                    # Get current smooth
                    curr_smooth = mesh.get_attribute(nc.smooth_attr_name)
                    # print 'smooth on', mesh.get_name(), curr_smooth

                    # Increment or not
                    if value:
                        new_smooth = curr_smooth +1
                        # print 'aggiungo'
                    else:
                        if curr_smooth != 0:
                            new_smooth = curr_smooth -1
                            # print 'sottraggo'
                        else:
                            new_smooth = curr_smooth

                    # print 'applico', new_smooth
                    mesh.set_attribute(nc.smooth_attr_name, new_smooth)


    #==================== COMPUTATION =======================
    def get_compute_state(self):
        '''
        get the compute state by looking only at the current attribute.
        Since we will never set the computation only on 1, the current compute state equals
        to the compute state of all the operators

        Returns:
            the current computation value
        '''
        return self.get_current_operator().get_network_object().get_attribute( nc.operator_compute_name )


    def set_compute_state(self, *values):
        '''
        Set the computation for the current network on or off

        Args:
            value: 0 for turning off computations.1 for turning on computations.

        Returns:
            None
        '''
        # Set the value of the compute on every network
        for operator in self.net_objs:
            operator.get_network_object().set_attribute(nc.operator_compute_name, values[0])

    #==================== VISIBILITY =======================
    def set_local_visibility(self, value, mesh_vis, curve_vis):
        """
        Set the local visibility option and run an update of the visibility status to update the maya scene.

        value: 		bool: 	True if showing only the current operators object, False if showing all
        visibility:	bool:	True / False state
        """

        # Change the visibility attribute
        self.local_visibility = value

        # # Once set, re update the visibility
        self.set_in_mesh_visibility(mesh_vis)
        self.set_curves_visibility(curve_vis)


    def get_objects_visibility(self):
        '''
        Return the visibility status of the operators involved in the current network.
        Return also the shading and lock state of the in mesh and result mesh.

        Args:
            None

        Returns:
            tuple:	in_shading:		bool
                    base_vis: 		bool
                    curves_vis:		bool
                    result_vis:		bool
                    result_lock: 	bool

        Raises:
            KeyError: Raises an exception.
        '''

        # Shading
        in_shading = self.get_current_operator().get_network_object().get_attribute( nc.shading_status )

        # Visibility
        base_vis = self.get_current_operator().get_network_object().get_attribute( nc.base_meshes_vis_name )
        curves_vis = self.get_current_operator().get_network_object().get_attribute( nc.curve_vis_name )
        result_vis = self.get_current_operator().get_network_object().get_attribute( nc.result_mesh_vis_name )

        # Lock
        result_lock = 1 if self.get_current_operator().get_result_mesh().get_attribute( 'overrideDisplayType' ) else 0


        return (in_shading, base_vis, curves_vis, result_vis, result_lock, self.local_visibility)


    def show_operator_result_for(self, input_ID):
        '''
        Show the result mesh only for the current specified ID,
        hide the other result meshes and turn off the computations for
        every network after the current to speed up the visualized editing.

        Args:
            None

        Returns:
            None
        '''
        for x, operator in enumerate(self.net_objs):
            # Save the net obj
            net_obj = operator.get_network_object()
            compute_val = net_obj.get_attribute(nc.operator_compute_name)

            if x == input_ID:
                operator.get_result_mesh().set_attribute('visibility', 1)

                if not compute_val:
                    net_obj.set_attribute(nc.operator_compute_name, 1)
            else:
                operator.get_result_mesh().set_attribute('visibility', 0)
                if x < input_ID:
                    net_obj.set_attribute(nc.operator_compute_name, 0)
                else:
                    if not compute_val:
                        net_obj.set_attribute(nc.operator_compute_name, 1)


    def set_in_mesh_visibility(self, *values):
        '''
        Set the visibility for the current involved input meshes

        Args:
            *visibility:

        Returns:
            None
        '''
        self.__update_visibility_in_current_mode(nc.base_meshes_vis_name, values[0])


    def set_curves_visibility(self, *values):
        '''
        Set the visibility for the current involved input meshes

        Args:
            *visibility:

        Returns:
            None
        '''
        self.__update_visibility_in_current_mode(nc.curve_vis_name, values[0])


    def set_out_mesh_visibility(self, *values):
        '''
        Set the visibility for the current involved input meshes

        Args:
            *visibility:

        Returns:
            None
        '''
        # Set the value of the visibility on every network
        for operator in self.net_objs:
            operator.get_network_object().set_attribute(nc.result_mesh_vis_name, values[0])


    #================= EXTRA - VISIBILITY =======================

    def set_out_mesh_lock(self, *values):
        '''
        Set the selection lock for the current result mesh

        Args:
            *visibility:

        Returns:
            None
        '''
        # Act only on the current network
        # operator = self.get_current_operator()

        # Set the result meshes mode to Normal or Reference
        display_type_value = 2 if values[0] else 0


        # Set the value of the visibility on every network
        for operator in self.net_objs:
            # Set the override on Normal or Reference
            operator.get_result_mesh().set_attribute('overrideDisplayType', display_type_value)
            operator.get_result_mesh().get_shape().set_attribute('overrideDisplayType', display_type_value)


    def set_in_mesh_shading(self, *values):
        '''
        Set the wireframe override on or off

        Args:
            *visibility:

        Returns:
            None
        '''

        # Set the value of the wireframe
        for operator in self.net_objs:
            # Set the override on Normal or Reference
            in_meshes = operator.get_input_meshes()

            for mesh in in_meshes:
                # Apply only to the clean 'in mesh' and not to result in mesh
                if mesh.get_object_type() == nc.hm_in_mesh:
                    # print ' setto l ovverride di %s a %s' % (mesh.get_shape().get_name(), values[0])
                    mesh.get_shape().set_attribute('overrideShading', values[0])

            # Track the change on the main group
            operator.get_network_object().set_attribute(nc.shading_status, values[0])


    #================ COLOR HILIGHTING =====================

    def set_in_mesh_color_hilighting(self, on_off):
        '''
        Args:
            on_off: Turn on or off the color for the in meshes

        Returns:
            None
        '''
        if on_off == 1:
            color = self.HILIGHTED_MESH_COLOR
        else:
            color = 0

        self.__set_color_hilighting( 'in_meshes', color )


    def set_curves_color_hilighting(self, on_off):
        '''
        Args:
            on_off: Turn on or off the color for the in meshes

        Returns:
            None
        '''
        if on_off == 1:
            color = self.HILIGHTED_CURVES_COLOR
        else:
            color = 0

        self.__set_color_hilighting( 'curves', color )

    #================ CLEANING =====================
    @util.undo
    def clean_current(self, keep_input):
        '''
        Delete all the hm nodes except the result of the current loaded operators network

        Args:
            keep_input: If on delete the input meshes after the cleaning has done

        Returns:
            None
        '''
        if self.net_objs:

            # Get the last result and set visible and selectable
            last_result = self.get_last_operator().get_result_mesh()
            last_result.set_attribute('visibility', 1)
            last_result.set_attribute('overrideEnabled', 0)
            last_result.get_shape().set_attribute('overrideEnabled', 0)
            last_result.remove_hm_attributes()
            last_result.get_shape().break_connection('inMesh', 'in')

            # Parent to the world
            cmds.parent(last_result.get_name(), world=True )


            # Delete all the related operators
            for operator in self.net_objs:
                # Force the restoring of the in meshes with this shitty thing
                operator.get_input_meshes()
                # Call the self delete method
                operator.delete(keep_input)

    @util.undo
    def clean_scene(self, keep_input):
        '''
        Delete all the hm nodes except the result of the current loaded operators network

        Args:
        keep_input: If on delete the input meshes after the cleaning has done

        Returns:
        None
        '''
        # Get all the compound from the scene
        compound_list =  cmds.ls(type='hmCompoundNode')


        while ( len(compound_list) != 0 ):

            # print 'compound_list: %s ' % compound_list

            # Init the first compound
            compound = hmNode.HmBaseObject( compound_list[0] )
            # print 'init from:', compound.get_name()


            # Get the result mesh
            to_choice = compound.get_connected('outputMesh', 'out', node_type = 'choice')

            if to_choice:
                # Get the mesh and init the handler with the current result
                choice = to_choice[0][0]
                result_mesh = choice.get_connected('output', 'out')[0][0]
                self.initialize_from_selection(result_mesh.get_name(), nc.hm_out_mesh)

                # print 'curr operator', self.get_current_operator().get_network_name(), 'init from ', result_mesh.get_name()
                # print 'in meshes ', self.get_current_operator().get_input_meshes()

                # This list stores the compound of the currently initialized network
                involved_compound = [operator.get_compound_node().get_name() for operator in self.net_objs]

                # Clean the current network of operators
                self.clean_current(keep_input)

                # print 'current network cleaned'

                # Remove from the loop list the compound that we have just delete
                for comp in involved_compound:
                    if comp in compound_list: compound_list.remove(comp)

            # No connections found, orphan node
            else:
                #Just delete the compound and move on
                compound.delete()
                compound_list.remove( compound.get_name() )



    #===========================================================
    #================== INTERNAL METHODS =======================
    #===========================================================
    def __disconnect_operator(self, oper_ID):
        """
        Disconnect the operator from the preview one and store connections data
        useful for reconnection operations.

                _______ Conn A >>>      ________ >>> Conn B
                |                       |
         _______|________        _______|________       ________________
        / Prev Operator /--  --/     Oper ID    /------/ Next Operator /
        ----------------      ------------------      -----------------

        Args:
            oper_ID: int: id of the operator to disconnect from the network.
  
        Return:
            orphan connections: [(preview_result_mesh, curr_result_mesh),
                                (prev_transform_connection, prev_shape_connection),
                                (curr_transform_connection, curr_shape_connection) ]

                                You will be able to use these connections to reconnect where these connections
                                where going to before this operaation.
        """
        return_data = { 'current_operator': {'mesh':None, 'connections':None},
                        'preview_operator': {'mesh':None, 'connections':None}  }


        net_lenght = len(self.net_objs) -1

        preview_operator = None
        current_operator = self.net_objs[oper_ID]

        # ================ DEFINE IF THERE IS A PREVIEW TO STORE ========================
        # Keeping in mind that the list is reversed..
        if oper_ID < net_lenght:
            preview_operator = self.net_objs[oper_ID+1]


        # ================ STORE CONNECTIONS DATA ==================================
        # Store current the out mesh connections
        # These connections will be actually good for reconnecting to its next operator
        curr_result_mesh = current_operator.get_result_mesh()
        curr_transform_connection = curr_result_mesh.get_connected('worldMatrix[0]', 'out')
        curr_shape_connection = curr_result_mesh.get_shape().get_connected('outMesh', 'out')

        # Update the return data
        #
        return_data['current_operator']['mesh'] = curr_result_mesh
        return_data['current_operator']['connections'] = (curr_transform_connection, curr_shape_connection)

        # ================ DISCONNECT CURRENT FROM NEXT ==================================
        # We disconnect the preview operator from the current.
        if curr_transform_connection:
            curr_result_mesh.break_connection('worldMatrix[0]', conn_type = 'out')
            curr_result_mesh.get_shape().break_connection('outMesh', conn_type = 'out')


        if preview_operator:
            # Store the preview mesh
            # Get the preview operator out mesh
            preview_result_mesh = preview_operator.get_result_mesh()

            # Store the out mesh connections
            prev_transform_connection = preview_result_mesh.get_connected('worldMatrix[0]', 'out')
            prev_shape_connection = preview_result_mesh.get_shape().get_connected('outMesh', 'out')


            # Update the return data
            #
            return_data['preview_operator']['mesh'] = preview_result_mesh
            return_data['preview_operator']['connections'] = (prev_transform_connection, prev_shape_connection)

            # ================ DISCONNECT PREVIEW FROM CURRENT ==================================
            # We disconnect the preview operator from the current.
            preview_result_mesh.break_connection('worldMatrix[0]', conn_type = 'out')
            preview_result_mesh.get_shape().break_connection('outMesh', conn_type = 'out')



        # ================ RETURN CONNECTIONS DATA ==================================
        return return_data


    def __connect_prev_with_next(self, curr_ID):
        """
        Given a valid ID, break its connections and reconnect
        its preview operator with its next operator.

        This is used for example in the delete where
        we will disconnect an operator anywhere in the chain
        for rejoining the two operators that had broken connections
        from the delete command.

        Args:
            :int 	curr_ID:	the ID for disconnection

        Return
            :[HmOperatorObject, connection data]  	:the removed object to be deleted or reallocated in the memory

        """
        # Disconnect the current operator from the networks
        #
        connections_data = self.__disconnect_operator(curr_ID)

        # Extract the preview mesh
        prev_mesh = connections_data['preview_operator']['mesh']

        # Use the connection of the current operator to connect the
        # preview operator with the next after the current
        mesh_connections = connections_data['current_operator']['connections']

        # And reconnect with each other
        #
        self.__connect_operators( prev_mesh, mesh_connections )


        # Update the internal memory
        #
        removed = self.net_objs[curr_ID]
        self.net_objs.remove(removed)


        return [removed, connections_data]


    def __connect_operators(self, hm_out_mesh, conn_list):
        """
        Connect the provided hm_out_mesh to the connection list.
        """


        transform_connection, shape_connection = conn_list

        # ================ CONNECT A - B ==================================
        # Only if we have more than one operator
        # Connect preview network to the new one, if connections are found...
        if transform_connection:
            for connection in transform_connection:
                hm_out_mesh.connect_attribute('worldMatrix[0]', connection[0], connection[1], force = True )
                if self.VERBOSE > 1:
                    print 'collego %s.%s a %s.%s' % (hm_out_mesh.get_name(), 'worldMatrix[0]', connection[0].get_name(), connection[1])
            for connection in shape_connection:
                hm_out_mesh.get_shape().connect_attribute('outMesh', connection[0], connection[1], force = True )
                if self.VERBOSE >1:
                    print 'collego %s.%s a %s.%s' % (hm_out_mesh.get_shape().get_name(), 'outMesh', connection[0].get_name(), connection[1])


    def __insert_operator(self, insert_after_ID):
        """
        Insert the operation in between two other operations by providing
        the id to insert.

                                                                  
         _______________        _______________
        /  Operator A  /--|---/   Operator B  /
        ----------------  |   ---------------
                          |
                     Insertion ID
                    ______|________
                   /  Operator C  /
                  ----------------


        """
        pass


    def __get_visibility_state_from_networks(self):
        """
        search for mismatch in visibility to determinate if
        we are in global or local visibility
        """
        # Check on all the networks for the in mesh to detect if we are in local
        base_vis = None
        self.local_visibility = 0

        for x, operator in enumerate(self.net_objs):

            # print 'LOOP', x, '================\n'
            current_v = operator.get_network_object().get_attribute( nc.base_meshes_vis_name )
            # print 'visibilita per %s' % operator.get_network_object().get_name(), current_v

            # If base vis has already a value from a preview iteration
            if base_vis != None:
                # print 'base vis esiste!'
                # and is different from the current, we are in local mode!
                if current_v != base_vis:
                    # print 'la visibilita corrente e diversa da quella precedente, LOCAL!'
                    self.local_visibility = 1
                    base_vis = current_v
                    break
                else:
                    # print 'tutto sembra essere in ALL'
                    base_vis = current_v

            else:
                # print 'prima iterazione...'
                base_vis = current_v

            # print 'LOOP', x, '================\n\n'

        # print 'FINE\n\n'


    def __update_visibility_in_current_mode(self, attribute_to_change, value):
        """
        This method implements the logic for turning on and
        off the elements based on the local or global visibility.

        Args:
            attribute_to_change: 	string: the name of the attribute on the current
                                            network/operator to change
            value:					int:	the value to set (usually is 0 / 1)
        """
        # print "SETTO VISIBILITA A", value
        # Set the value of the visibility on every network
        for operator in self.net_objs:

            curr_oper_obj = operator.get_network_object()

            if self.local_visibility:
                # print 'LOCAL!'
                # If we have to show the meshes, turn on only the current, and turn off all the others
                if value == 1:
                    # print 'SHOW!'
                    # Set the value on the current one
                    if operator == self.get_current_operator():
                        curr_oper_obj.set_attribute(attribute_to_change, 1)
                    else:
                        curr_oper_obj.set_attribute(attribute_to_change, 0)

                # If we have to hide them, hide them all
                if value == 0:
                    # print 'HIDE'
                    curr_oper_obj.set_attribute(attribute_to_change, 0)

            # If we are in global visivility, do the same on everything
            else:
                curr_oper_obj.set_attribute(attribute_to_change, value)


    def __set_color_hilighting(self, operate_on, color):
        '''
        Set the color hilighting to the specified elements

        Args:
            operate_on: string 'in_meshes' or 'curves' describing on what we need to operate
            color: integer index of the color

        Returns:
            None
        '''
        def set_color_on_shape(in_objects, color_ID):
            '''
            Do the actual set attributes on every shape

            Args:
                in_objects: list of HmBaseObjects to operate on
                color_ID: integer index of the color
            '''
            for el in in_objects:
                # print 'elemento', el.get_name()
                shape = el.get_shape()
                # If we need to turn it off simply set it to 0 Since
                # we need to keep the override always On for the mesh
                shape.set_attribute('overrideEnabled', 1)
                shape.set_attribute('overrideColor', color_ID)

        if self.net_objs:
            for x,operator in enumerate(self.net_objs):
                # Deactivate all the object colors not
                # in the current operator
                if x != self.current_obj:
                    color_to_set = 0
                else:
                    color_to_set = color

                if operate_on == 'in_meshes':
                    set_color_on_shape( operator.get_input_meshes(), color_to_set )
                elif operate_on == 'curves':
                    set_color_on_shape( operator.get_offset_curves(), color_to_set )
                else:
                    raise Exception( '%s is not a valid category for the color hilighting feature' % operate_on )


